# WTFPL

import subprocess
import shlex
import shutil
import os
import re
import signal
import sys


class MonitoredApp:
    def __init__(self,name,options):
        self.name = name
        self.options = options
        if "shell" in options and options["shell"]:
            self.cmd_tokens = options["cmd"]
        else:
            self.cmd_tokens = shlex.split(options["cmd"])
        self.persist = options["persist"] if "persist" in options else False
        self.proc = None
        self.returnValue = None

    @staticmethod
    def get(name, options):
        if "acquire" in options and options["acquire"]:
            aa = AcquiredApp(name,options)
            if aa.getpid():
                return aa
            #else
            aa.terminateIfNonpersistent()
        return OwnApp(name,options)

    def terminateIfNonpersistent(self):
        if not self.persist:
            self.terminate()

    def isAquired(self):
        return False

class OwnApp(MonitoredApp):
    def __init__(self,name,options):
        MonitoredApp.__init__(self,name,options)

    def launch(self):
        if not self.proc or not self.running():
            options = {}
            if "cwd" in self.options:
                options["cwd"] = self.options["cwd"]
            if "shell" in self.options and self.options["shell"]:
                options["shell"] = True
                options["executable"] = "/bin/bash"
            if "detach" in self.options and self.options["detach"]:
                options["preexec_fn"]=os.setpgrp

            self.proc = subprocess.Popen(self.cmd_tokens, **options)
            return self

    def terminate(self):
        if self.proc:
            self.proc.terminate()
            try:
                self.proc.wait(1)
            except subprocess.TimeoutExpired as ex:
                print("do not terminate, killing")
                self.proc.kill()
                self.proc.wait(1)
            self.returnValue = self.proc.poll()
            self.proc = None

    def running(self):
        if not self.proc:
            return False
        else:
            return self.proc.poll() is None

    def poll(self):
        return self.proc.poll() if self.proc else self.returnValue


class AcquiredApp(MonitoredApp):
    def __init__(self,name,options):
        MonitoredApp.__init__(self,name,options)
        self.pid = self.getpid()
        if shutil.which("strace") is None:
            print("acquire flag require strace to be installed.")
            sys.exit(-1)

        strace_cmd = "strace -e none -e exit_group -p "+str(self.pid)
        self.strace = subprocess.Popen(shlex.split(strace_cmd),stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf-8")
        self.straceOut = None
        self.straceEnded = False
    def isAquired(self):
        return True

    def getpid(self):
        return self.getPidBasedOnGrepCmd()

    def getPidBasedOnGrepCmd(self):
        grepCommand = """ps -eo pid,cmd |egrep "^ *[0-9]* """ \
                + self.options["cmd"] + """$" --color|awk '{print $1;}'"""
        output = subprocess.check_output(grepCommand, shell=True)
        lines = output.decode("utf8").strip().split("\n")
        if len(lines) >1:
            print("warning: multiple pid found")
        if len(lines) == 1 and len(lines[0]) >0:
            return int(lines[0])
        else:
            return None

    def launch(self):
        if self.straceEnded:
            oa = OwnApp(self.name,self.options)
            oa.launch()
            return oa
        else:
            return self

    def running(self):
        return self.strace.poll() is None

    def poll(self):
        if self.running():
            return None
        else:
            if not self.straceEnded:
                #first poll after strace ended
                self.straceEnded = True
                outs = self.strace.stdout.read()
                # print("read outs:"+str(outs))
                errs = self.strace.stderr.read()
                # print("read errs:"+str(errs))
                match = re.search(r"\+\+\+ exited with (\d+) \+\+\+",errs)
                if match:
                    self.straceOut = int(match.group(1))
                else:
                    self.straceOut = -1
            return self.straceOut


    def terminate(self):
        if not(self.straceEnded):
            os.kill(self.pid, signal.SIGTERM)
            self.strace.wait(.5)
            self.poll()

    def terminateIfNonpersistent(self):
        super().terminateIfNonpersistent()
        if self.strace.poll() is None:
            self.strace.terminate()


# WTFPL

def main():
    import toml

    import argparse

    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--config", help="Config file",default="lambada.toml")
    parser.add_argument("--noauto", help="does not launch tasks", action="store_false", dest="auto")

    args = parser.parse_args()

    config = toml.load(args.config)
    print(config)

    from lambada.MonitoredApp import MonitoredApp
    from lambada.GUI import App

    a = App(100)
    for s in config:
        ma = MonitoredApp.get(s,config[s])
        a.addApplication(ma)
        if args.auto:
            ma.launch()

    a.mainloop()
    a.stopAll()

if __name__=="__main__":
    main()



# WTFPL

import tkinter as tk


class App (tk.Tk):
    def __init__(self, refresh_rate = 1000):
        tk.Tk.__init__(self)
        self.title("Launcher")
        tk.Label(self,text="Launch panel",justify="center").grid(row=0,columnspan=4)
        self.lines = []
        self.refresh_rate = refresh_rate
        self.after(self.refresh_rate,self.checkProcesses)

        for i in range(4):
            self.columnconfigure(i, weight=1)

    def addApplication(self,ma):
        idx = len(self.lines) + 1
        self.lines.append(Line(self,ma,idx))



    def checkProcesses(self):
        for line in self.lines:
            line.check()
        self.after(self.refresh_rate,self.checkProcesses)

    def stopAll(self):
        for line in self.lines:
            line.quitApp()

class Line:
    # does not inherit Frame because grid layout is shared with parent
    # might switch to pack() instead
    def __init__(self,master,ma,row):
        self.master = master
        self.ma = ma
        self.row = row
        self.l = tk.Label(master,text=ma.name)
        self.l.grid(row=row,column=0)
        self.updateLabelColor()
        self.c = tk.Canvas(master,width=20,height=20,bg="purple")
        self.c.grid(row=row,column=1,sticky="news")
        self.buttonRun = tk.Button(master,text="▶",command=self.runApp)
        self.buttonRun.grid(row=row,column=2,sticky="news")
        self.buttonStop = tk.Button(master,text="■",command=self.stopApp)
        self.buttonStop.grid(row=row,column=3,sticky="news")
        self.buttonRestart = tk.Button(master,text="↺",command=self.restartApp)
        self.buttonRestart.grid(row=row, column=4, sticky="news")

        self.returnStatus = StatusLight(master)
        self.returnStatus.grid(row=row, column=5, sticky="news")

        master.rowconfigure(row,weight=1)

    def runApp(self):
        self.returnStatus.setColor("grey")
        self.master.after(0,self.doRunApp)

    def doRunApp(self):
        self.ma = self.ma.launch()
        self.updateLabelColor()

    def stopApp(self):
        self.ma.terminate()
    def restartApp(self):
        self.stopApp()
        self.master.after(10,self.runApp())
    def quitApp(self):
        self.ma.terminateIfNonpersistent()
    def check(self):
        if self.ma.running():
            self.c.config(bg="green")
            self.buttonRun.config(state=tk.DISABLED)
            self.buttonStop.config(state=tk.NORMAL)
            self.returnStatus.setColor("grey")
        else:
            self.c.config(bg="red")
            self.buttonRun.config(state=tk.NORMAL)
            self.buttonStop.config(state=tk.DISABLED)
            self.returnStatus.setColor("green" if self.ma.poll() == 0 else "red")

    def updateLabelColor(self):
        self.l["fg"] = "blue" if self.ma.isAquired() else "black"


class StatusLight(tk.Canvas):
    def __init__(self,master,**args):
        args["height"] = args["height"] if "height" in args else 20
        args["width"] = args["width"] if "width" in args else 20
        args["bg"] = args["bg"]  if "bg" in args else master["bg"]
        tk.Canvas.__init__(self,master,**args)
        self.o = self.create_oval(5,5,15,15,fill="grey",outline="black")
    def setColor(self,color):
        self.itemconfig(self.o,fill = color)
